# Copyright 2019 The Forte Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from forte.data.readers.conll03_reader import *
from forte.data.readers.conllu_ud_reader import *
from forte.data.readers.ms_marco_passage_reader import *
from forte.data.readers.deserialize_reader import *
from forte.data.readers.html_reader import *
from forte.data.readers.multipack_sentence_reader import *
from forte.data.readers.multipack_terminal_reader import *
from forte.data.readers.ontonotes_reader import *
from forte.data.readers.plaintext_reader import *
from forte.data.readers.prodigy_reader import *
from forte.data.readers.race_multi_choice_qa_reader import *
from forte.data.readers.string_reader import *
from forte.data.readers.sem_eval_task8_reader import *
from forte.data.readers.sst2_reader import *
from forte.data.readers.openie_reader import *
from forte.data.readers.ag_news_reader import *
from forte.data.readers.largemovie_reader import *
from forte.data.readers.misc_readers import *
from forte.data.readers.classification_reader import *
from forte.data.readers.audio_reader import *
